/*****************************************************************************
/ Assignment 3 Part 2: Application                                          
/ Bounded Buffer problem.
/ Description:  Producer produces strings and put it in a bounded buffer
/               Consumer takes strings out of the buffer and prints them.
/               Buffer can contain 10 strings at most at anytime.
/****************************************************************************/


#include <stdlib.h>
#include <stdio.h>
#include <time.h>
#include <string.h>
#include "minithread.h"
#include "synch.h"

#define STR_LENGTH 10
#define BUFFER_LENGTH 10
#define NUM_ITERATIONS 1000

semaphore_t full, empty;

volatile char *buffer[BUFFER_LENGTH+1];
volatile int front, rear;

static int producer()
{
  int count, i;
  char string[STR_LENGTH + 1];


  for (i = 0; i<NUM_ITERATIONS; i++) {

    /* randomly generate strings of length STR_LENGTH */
    for(count = 0; count < STR_LENGTH; count++)
      string[count] = random() % 26 + 'A';

    string[STR_LENGTH] = '\0'; /* Null terminated string */


    /* Counting semaphore.  Proceed only when queue is not full*/
    /* Take one empty */
    semaphore_P(empty);
    /* Binary semaphore.  Atomic access to the buffer queue*/

    /* Put the generated string on the queue*/
	/* Note we don't need to synchronize access to the buffer
	   Even if both the threads access it at the same time, 
	   nothing bad will happen */
	buffer[rear] = strdup(string);
	rear = (rear+1) % (BUFFER_LENGTH+1);
    fprintf(stderr, "String %s is put into buffer.\n", string);
    fflush(stderr);

    /* create one full*/
    semaphore_V(full);
  }
		
  return 0;
}


static int consumer()
{
  int i;
  char* string;
  
  for (i = 0; i<NUM_ITERATIONS; i++) {

    /* Counting semaphore. Proceed only if queue !empty*/
    /* Take one full*/
    semaphore_P(full);

    /* Remove one string from queue and print*/
	string = buffer[front];
	front = (front+1) % (BUFFER_LENGTH+1);
    fprintf(stdout, "string %s is consumed.\n", string);
    fflush(stdout);
	free(string);        /* Free this string created by strdup() */

    /* create one empty(*/
    semaphore_V(empty);
  }

   return 0;
}

    
static int test()
{
  srandom(time(NULL));
  
  full = semaphore_create();
  semaphore_initialize(full, 0); /* counting semaphore*/

  empty = semaphore_create();
  semaphore_initialize(empty, BUFFER_LENGTH); /* counting semaphore*/

  minithread_fork((proc_t)&producer, NULL);
  minithread_fork((proc_t)&consumer, NULL);

  return 0;
}

int main(void)
{
  minithread_system_initialize(test, NULL);
}


